// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.

#include <eopcc/cloud.hpp>
#include <eopcc/utility.hpp>
#include <eopcc/utm_zone.hpp>

namespace eopcc {

void PointCloud::append(PointCloud& other) {
	loc.insert(loc.end(), other.loc.begin(), other.loc.end());
	intensity.insert(intensity.end(), other.intensity.begin(), other.intensity.end());


	if (has_color && other.has_color) {
		color.insert(color.end(), other.color.begin(), other.color.end());
	}

	if (has_time && other.has_time) {
		gps_time.insert(gps_time.end(), other.gps_time.begin(), other.gps_time.end());
	}


	if (has_las_metadata && other.has_las_metadata) {
		las_metadata.insert(las_metadata.end(), other.las_metadata.begin(), other.las_metadata.end());
	}

	// resizes any attributes we had to skip
	resize(size());
}

void PointCloud::set_from(size_t index, const PointCloud& other, int other_index) {
	loc[index] = other.loc[other_index];
	intensity[index] = other.intensity[other_index];

	if (has_las_metadata && other.has_las_metadata) {
		las_metadata[index] = other.las_metadata[other_index];
	}

	if (has_color && other.has_color) {
		color[index] = other.color[other_index];
	}

	if (has_time && other.has_time) {
		gps_time[index] = other.gps_time[other_index];
	}
}


PointCloud PointCloud::subcloud(size_t sub_start, size_t sub_end) const {
	PointCloud out_cloud;
	out_cloud = (*this);

	size_t num_valid = sub_end - sub_start;

	out_cloud.resize(num_valid);

	for (size_t in_index=sub_start; in_index < sub_end; ++in_index) {
		out_cloud.set_from(in_index - sub_start, (*this), in_index);
	}

	return out_cloud;
}

Point PointCloud::mean() const {
	Point mean;
	mean.x = 0;
	mean.y = 0;
	mean.z = 0;

	for (const Point& pt : loc) {
		mean = mean + pt;
	}

	mean.x /= loc.size();
	mean.y /= loc.size();
	mean.z /= loc.size();


	return mean;
}

void PointCloud::translate(const Point& translation) {
	for (Point& pt : loc) {
		pt = pt + translation;
	}

	extent = extent + translation;
}




Extent PointCloud::calculate_extent() const {
	pos_type_t x_min = std::numeric_limits<pos_type_t>::max();
	pos_type_t y_min = std::numeric_limits<pos_type_t>::max();
	pos_type_t z_min = std::numeric_limits<pos_type_t>::max();
	pos_type_t x_max = -std::numeric_limits<pos_type_t>::max();
	pos_type_t y_max = -std::numeric_limits<pos_type_t>::max();
	pos_type_t z_max = -std::numeric_limits<pos_type_t>::max();
	for (size_t i=0; i < loc.size(); ++i) {
		if (loc[i].x < x_min) {
			x_min = loc[i].x;
		}
		if (loc[i].y < y_min) {
			y_min = loc[i].y;
		}
		if (loc[i].z < z_min) {
			z_min = loc[i].z;
		}

		if (loc[i].x > x_max) {
			x_max = loc[i].x;
		}
		if (loc[i].y > y_max) {
			y_max = loc[i].y;
		}
		if (loc[i].z > z_max) {
			z_max = loc[i].z;
		}
	}

	Extent t_extent;

	t_extent.min_x = x_min;
	t_extent.max_x = x_max;
	t_extent.min_y = y_min;
	t_extent.max_y = y_max;
	t_extent.min_z = z_min;
	t_extent.max_z = z_max;

	return t_extent;
}


PointCloud RandomCloud(size_t number_points, const Extent& extent) {
    std::random_device rd;
    std::mt19937 gen(rd());

    PointCloud output;
    output.extent = extent;
    output.resize(number_points);

    for (size_t i=0; i < number_points; ++i) {
        output.loc[i] = extent.random_sample(gen);
    }

    return output;
}

int PointCloud::inside_extent(const Extent& extent) {
	int count = 0;

	for (Point& pt : loc) {
		if (extent.contains(pt)) {
			count++;
		}
	}

	return count;
}


// Given a vector of keep/trash flags creates a new vector with only the keep members
PointCloud PointCloud::filter(const std::vector<bool>& filter_vector) const {
    PointCloud out_cloud;

    // This is the best safe way to do this without
    // screwing myself when I add more members to this class
    // (But it uses a little more memory that necessary)
    out_cloud = (*this);


    size_t out_index = 0;
    size_t num_valid = 0;
    for (size_t in_index=0; in_index < filter_vector.size(); ++in_index) {
        if (filter_vector[in_index])
            num_valid += 1;
    }

    out_cloud.resize(num_valid);

    for (size_t in_index=0; in_index < filter_vector.size(); ++in_index) {
        if (filter_vector[in_index]) {
            out_cloud.set_from(out_index, (*this), in_index);
            out_index++;
        }
    }

    return out_cloud;
}

void PointCloud::filter_inplace(const std::vector<bool>& filter_vector) {
    size_t out_index = 0;
    size_t num_valid = 0;
    for (size_t in_index=0; in_index < filter_vector.size(); ++in_index) {
        if (filter_vector[in_index])
            num_valid += 1;
    }

    for (size_t in_index=0; in_index < filter_vector.size(); ++in_index) {
        if (filter_vector[in_index]) {
            set_from(out_index, (*this), in_index);
            out_index++;
        }
    }

    resize(num_valid);
}

int PointCloud::inside_extent(const PolygonExtent& extent) {
		int count = 0;

		for(Point& pt : loc) {
			if (extent.contains(pt)) {
				count++;
			}
		}

		return count;
}

} // eopcc namespace end
